/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: complex.cpp,v 1.3 2001/08/09 11:37:54 cxsc Exp $ */

#include "complex.hpp"
#include "dot.hpp"
#include "cdot.hpp"
#include "real.hpp"
#include "rmath.hpp"
#include "idot.hpp"
#include "interval.hpp"

namespace cxsc {

complex::complex(const cdotprecision & a) throw()
{
   *this=rnd(a);
}

complex & complex::operator =(const cdotprecision & a) throw()
{
   return *this=rnd(a);
}

bool operator== (const complex & a, const dotprecision & b)    throw() { return !a.im && a.re==b; }
bool operator== (const dotprecision & a, const complex & b)    throw() { return !b.im && a==b.re; }
bool operator!= (const complex & a, const dotprecision & b)    throw() { return !!a.im || a.re!=b; }
bool operator!= (const dotprecision & a, const complex & b)    throw() { return !!b.im || a!=b.re; }

complex operator *(const complex &a,const complex &b) throw()
{
   complex tmp;
    
   dotakku[4] = 0.0;
   accumulate (dotakku[4],  a.re, b.re);
   accumulate (dotakku[4], -a.im, b.im);
   rnd (dotakku[4], tmp.re, RND_NEXT);

   dotakku[4] = 0.0;
   accumulate (dotakku[4], a.re, b.im);
   accumulate (dotakku[4], a.im, b.re);
   rnd (dotakku[4], tmp.im, RND_NEXT);

   return tmp;
}

static const int minexpo = -1021, maxexpo = 1022,
                 minexpom = -914, maxexpo1 = 1022;

void product(real a,real b,real c,real d,
              int & overfl,real & p1,interval & p2)
{
   a_intg exa, exb, exc, exd;  // Exp. von a-d
   dotprecision dot;
   int inexact;

   overfl  = 0;
   inexact = 0; // false

   dot = 0.0;
   exa = expo(a);
   exb = expo(b);
   exc = expo(c);
   exd = expo(d);

   if ( sign(a) == 0  ||   sign(b) == 0 )    //  a == b == 0.0
      if ( sign(c) == 0  ||  sign(d) == 0 )  //  a == b == c == d == 0
                                             //     dot := #(0);
         ;    // No Operation necessary
      else 
      {
         //  a == b == 0, c,d != 0
         if (exc+exd > maxexpo1) 
         {
            //  overflow !
            c = comp( mant(c), exc-maxexpo );
            overfl = 1;
         } else 
         if  (exc+exd < minexpom) 
         {
            // minexpom, da Unterlauf auch durch Ausloeschung moeglich ist
            // underflow !
            c = comp( mant(c), exc+maxexpo );
            overfl = -1;
         }
         accumulate(dot,c,d);
      } else //  a,b != 0
      if ( sign(c) == 0  ||  sign(d) == 0 ) 
      {
         //  a,b != 0, c == d == 0
         if (exa+exb > maxexpo1) 
         {
            //  overflow !
            a = comp( mant(a), exa-maxexpo );
            overfl = 1;
         } else 
         if (exa+exb < minexpom) 
         {
            // minexpom, da Unterlauf auch durch Ausloeschung moeglich ist
            // underflow !
            a = comp( mant(a), exa+maxexpo );
            overfl = -1;
         }
         accumulate(dot,a,b);
      } else 
      {
         // a,b,c,d != 0
         if (exa+exb > maxexpo1) 
         {
            //  overflow bei a*b
            a = comp( mant(a), exa-maxexpo );
            if (exc>0)
               c = comp( mant(c), exc-maxexpo );
            else if (exd>0)
               d = comp( mant(d), exd-maxexpo );
            else 
            {
               // underflow wegen Skalierung bei c*d
               c = 0.0;
               inexact = 1; // true
            }
         } else if (exc+exd > maxexpo1) 
         {
            // overflow bei c*d
            c = comp( mant(c), exc-maxexpo );
            if (exa>0)
               a = comp( mant(a), exa-maxexpo );
            else if (exb>0)
               b = comp( mant(b), exb-maxexpo );
            else 
            {
               // underflow wegen Skalierung bei a*b
               a = 0.0;
               inexact = 1; // true
            }
            overfl = 1;
         } else 
         if ( exa+exb < minexpom  &&  exc+exd < minexpom ) 
         {
            //  underflow bei a*b und bei c*d
            a = comp( mant(a), exa+maxexpo );
            c = comp( mant(c), exc+maxexpo );
            overfl = -1;
         }
         accumulate(dot, a, b);
         accumulate(dot, c, d);
      }
      
   p1 = rnd(dot);
   dot -= p1;
   p2 = rnd(dot);

   if (inexact)
      p2 = _interval(pred(Inf(p2)),succ(Sup(p2)));

}


real quotient (real z1, interval z2, real n1, interval n2, int round,
               int zoverfl, int noverfl)
{
   real q1, scale;
   interval q2, nh;
   idotprecision id;
   int vorz, anz_scale;

   vorz = sign(z1);

   if ( zoverfl == -1  &&  noverfl==1 ) 
   {
      //  Ergebnis liegt im Unterlaufbereich
      switch (round) 
      {
         case RND_DOWN:
            if (vorz>=0) 
               q1 = 0.0;
            else         
               q1 = -MinReal;
            break;
         case RND_NEXT:
            q1 = 0.0;
            break;
         case RND_UP:
            if (vorz<=0) 
               q1 = 0.0;
            else         
               q1 = MinReal;
            break;
      } // switch
   } else if ( zoverfl==1  &&  noverfl==-1 ) 
   {
      //  Ergebnis liegt im Ueberlaufbereich
      switch (round) 
      {
         case RND_DOWN:
            if (vorz>=0) 
               q1 = MaxReal;
            else         
               q1 = -MaxReal-MaxReal; // Overflow
            break;
         case RND_NEXT:
            q1 = MaxReal+MaxReal;       // Overflow
            break;
         case RND_UP:
            if (vorz<=0) 
               q1 = -MaxReal;
            else         
               q1 = MaxReal+MaxReal; // Overflow
            break;
      } // switch
   } else 
   {
      q1 = divd(z1, n1);  // down, damit q2 >=0 wird
      nh = _interval(addd(n1, Inf(n2)),
                   addu(n1, Sup(n2)));

      // q2:= ##( z1 - q1*n1 + z2 - q1*n2 );
      id = z1;
      accumulate(id, -q1, n1);
      id += z2;
      accumulate(id, -q1, n2);
      q2 = rnd(id);

      switch (round) 
      {
         case RND_DOWN:
            q1 = adddown(q1, divd(Inf(q2), Sup(nh)));
            break;
         case RND_NEXT:
            q1 = q1 + (Inf(q2)+Sup(q2))*0.5/n1;
            break;
         case RND_UP:
            q1 = addup(q1, divu(Sup(q2), Inf(nh)));
            break;
      } // switch


      //  Rueckskalierung, falls Zaehler/Nenner - Over/Underflow :

      //  Eigentlich wie folgt:

      //  q1:= comp( mant(q1), expo(q1) + (zoverfl-noverfl)*maxexpo );
      //  Da aber obiges bei zoverfl-noverfl < 0 unkontrollierten underflow
      //  erzeugen kann ( mit Fehlermeldung 'inexact' vom Laufzeitsystem ),
      //  wird gerichtet gerundet zurueck skaliert :

      anz_scale = zoverfl - noverfl;
      if (anz_scale > 0)
         scale = comp( mant(1.0), expo(1.0)+maxexpo );
      else if (anz_scale < 0)
         scale = comp( mant(1.0), expo(1.0)-maxexpo );
      else 
         scale = 1.0;
      
      if (anz_scale < 0) 
         anz_scale = -anz_scale;

      switch (round) 
      {
         case RND_DOWN:
            q1 = multdown(q1, scale);
            if (anz_scale==2)
               q1 = multdown(q1, scale);
            break;
         case RND_NEXT:
            q1 = q1 * scale;
            if (anz_scale==2)
               q1 = q1 * scale;
            break;
         case RND_UP:
            q1 = multup(q1, scale);
            if (anz_scale==2)
               q1 = multup(q1, scale);
            break;
      } // switch

   }

   return q1;
}



complex _c_division(complex a, complex b, int round)
{
   int zoverflow, noverflow;
   real z1, n1;
   interval z2, n2;
   complex tmp;

   product (Re(b), Re(b), Im(b), Im(b), noverflow, n1, n2);
   product (Re(a), Re(b), Im(a), Im(b), zoverflow, z1, z2);
   SetRe (tmp, quotient (z1, z2, n1, n2, round, zoverflow, noverflow));
   product (Im(a), Re(b), -Re(a), Im(b), zoverflow, z1, z2);
   SetIm (tmp, quotient (z1, z2, n1, n2, round, zoverflow, noverflow));
   return tmp;
}

complex c_divn (const complex & a, const complex & b)  
{
   return _c_division(a, b, RND_NEXT);
}

complex c_divd (const complex & a, const complex & b)  
{
   return _c_division(a, b, RND_DOWN);
}

complex c_divu (const complex & a, const complex & b)  
{
   return _c_division(a, b, RND_UP);
}

complex operator /(const complex &a,const complex &b) throw()
{
   // if(!b) throw DIVZERO
   return c_divn(a,b);
}

real abs2(const complex &a) throw()
{
   dotakku[4]=0;
   accumulate(dotakku[4],a.re,a.re);
   accumulate(dotakku[4],a.im,a.im);
   return rnd(dotakku[4]);
}

const int grenzexpo = 500;

real abs (complex z) throw()
{
   real hilf, zRE, zIM;
   int scale;

   zRE = Re(z);
   zIM = Im(z);

   // ***   ABS:= SQRT( SQR( z.RE ) + SQR( z.IM ) )   ***

   if ( expo(zRE) >= grenzexpo  ||  expo(zIM) >= grenzexpo ) 
   {
      //  Ueberlauf vermeiden
      SetRe(z, comp( mant(zRE), expo(zRE)-grenzexpo )); // Unterlauf durch diese
      SetIm(z, comp( mant(zIM), expo(zIM)-grenzexpo )); // Skalierung moeglich !
      scale = grenzexpo;
   } else if ( expo(zRE) <= -grenzexpo+1  &&  expo(zIM) <= -grenzexpo+1 ) 
   {
      //  Unterlauf vermeiden
      SetRe(z, comp( mant(zRE), expo(zRE)+grenzexpo ));
      SetIm(z, comp( mant(zIM), expo(zIM)+grenzexpo ));
      scale = -grenzexpo;
   } else
      //  Kein Grund zur Panik
      scale = 0;

   hilf = sqrt(abs2(z)); // Genauigkeit ?????
   return comp( mant(hilf), expo(hilf)+scale );
}



// ---- Ausgabefunkt. ---------------------------------------

std::ostream & operator << (std::ostream &s, const complex& a) throw()
{
   s << '('          
     << a.re << ','  
     << a.im       
     << ')';
   return s;
}
std::string & operator << (std::string &s, const complex &a) throw()
{
   s+='(';
   s << a.re;
   s+=',';
   s << a.im; 
   s+=')';
   return s;
}

std::istream & operator >> (std::istream &s, complex &a) throw()
{
   char c;

   skipeolnflag = inpdotflag = true;
   c = skipwhitespacessinglechar (s, '(');
   if (inpdotflag) 
      s.putback(c);

   s >> a.re;

   skipeolnflag = inpdotflag = true;
   c = skipwhitespacessinglechar (s, ',');
   if (inpdotflag) s.putback(c);

   s >> a.im >> RestoreOpt;

   if (!waseolnflag) 
   {
      skipeolnflag = false, inpdotflag = true;
      c = skipwhitespaces (s);
      if (inpdotflag && c != ')') 
         s.putback(c);
   }

   /*if (a.re > a.im) {
      errmon (ERR_INTERVAL(EMPTY));
   } */         
   return s;
}

std::string & operator >> (std::string &s, complex &a) throw()
{
   s = skipwhitespacessinglechar (s, '(');
   s >> SaveOpt >> RndDown >> a.re;
   s = skipwhitespacessinglechar (s, ',');
   s >> RndUp >> a.im >> RestoreOpt;
   s = skipwhitespaces (s);

   if (s[0] == ')') 
      s.erase(0,1);

    /*if (a.re > a.im) {
      errmon (ERR_INTERVAL(EMPTY));
    } */                         
   return s;
}

void operator >>(const std::string &s,complex &a) throw()
{
   std::string r(s);
   r>>a;
}
void operator >>(const char *s,complex &a) throw()
{
   std::string r(s);
   r>>a;
}

} // namespace cxsc
