/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: l_imath.cpp,v 1.4 2001/08/09 12:52:39 cxsc Exp $ */

#include <math.h>
#include "l_imath.hpp"
#include "imath.hpp"
#include "rmath.hpp"

namespace cxsc {

#define  CXSC_One       1.
#define  CXSC_Zero      0.
#define  CXSC_MinusOne  -1.

l_interval pow(const l_interval & x, const l_interval & e) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)
{
   int         stagsave = stagprec,
               stagmax = 19,
               intexp;
               
   bool        fertig;

   l_interval  y;
   interval    dx = _interval(x),
               de = _interval(e),
               einfachgenau;
   real        supabsde = Sup(abs(de));

   einfachgenau = pow(dx,de);

   fertig = false;
   if (Inf(de) == Sup(de)) // &&
      if (supabsde < 32768.0) 
      {
         intexp = int(_double(_real(Sup(e))));
         if (_real(intexp) == Sup(e)) 
         {
            y = power(x,intexp);   // Integer-Potenz wesentlich schneller
            fertig = true;
         }
      }
      
   if (!fertig) 
   {
      if (Inf(x) < 0.0) 
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval pow(const l_interval & x, const l_interval & e)"));
      else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_One)
         y = x;
      else if (Inf(de) == Sup(de) && Sup(de) == CXSC_One)
         y = x;
      else if (Inf(de) == Sup(de) && Sup(de) == CXSC_Zero)
         y = 1.0;
      else 
      {
         if (stagprec < stagmax) 
            stagprec++;
         else                    
            stagprec = stagmax;
         y = exp(e*ln(x));
         stagprec = stagsave;
         y = adjust(y);
         y = y & einfachgenau;
      }
   }

   return y;
}

l_interval power(const l_interval &x, int n)       // Power(x,n)
{
   int         stagsave = stagprec,
               stagmax = 19;
              
   bool        neg = false;
              
   long int    zhi = 2;
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  y, neu;

   einfachgenau = power(dx,n);

   if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_One)
      y = x;
   else if (n == 0)
      y = adjust(_l_interval(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      
      if (n == 1)
         y = x;
      else if (n == 2)
         y = sqr(x);
      else 
      {
         if (n < 0) 
         {
            neg = true;
            n = -n;
         }
         // Initialisierung
         if (n%2)  
            y = x;
         else      
            y = _l_interval(1.0);  // Przision wird bei 1 Mult. auf
                                        // aktuellen Wert gesetzt;
         // Berechnugn durch binre Darstellung der n
         neu = x*x;
         do {
            if ((n/zhi)%2)  y *= neu;
            zhi += zhi;
            if (zhi <= n)  // letzte Mult. entfaellt --> schneller
               neu *= neu;
         } while (zhi <= n);

         if (neg) 
            y = 1.0/(y);
      }
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval sqr(const l_interval & x)              // Sqr(x)
{
   l_interval	y;

   if (Inf(x) >= 0.0)       /* result = [x.inf*x.inf, x.sup*x.sup] */
      y = x * x;
   else if (Sup(x)<= 0.0) 
   { 
      /* result = [x.sup*x.sup, x.inf*x.inf] */
      y = _l_interval (-Sup(x) , -Inf(x));
      y = (y) * (y);
   } else 
   {                   
      /* result = [0.0, max(x.sup*x.sup,x.inf*x.inf)] */
      if (abs(Inf(x)) >= abs(Sup(x)))
         y = _l_interval(0.0, abs(Inf(x)));
      else
         y = _l_interval(0.0, abs(Sup(x)));
      y = (y) * (y);
   }
	return y;
}

l_interval sqrt(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)            // Sqrt(x)
{
   int         stagsave = stagprec,
               stagmax = 19,
               stagsave2;
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  a1, y, t, mt;

   einfachgenau = sqrt(dx);

   if (Inf(x) < 0.0) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval sqrt(const l_interval & x)"));
   else if (Inf(dx) == Sup(dx) && (Sup(dx) == CXSC_Zero || Sup(dx) == CXSC_One))
      y = x;
   else 
   {
      // Einsatz des Intervall-Newton-Verfahrens y = m(y)-f(m(y))/f'(y)
      t = sqrt(_interval(x));
      if (stagprec < stagmax) 
         stagsave2 = stagprec+1;
      else                    
         stagsave2 = stagmax;
      stagprec = 1;
      while (stagprec < stagsave2) 
      {
         stagprec += stagprec;
         if (stagprec > stagmax) 
            stagprec = stagmax;
         //      mt = Inf(t)+(Sup(t)-Inf(t))*0.5;
         mt = mid(t);
         t = mt-((mt*mt-x)/(2.0*t));
      }
      stagprec = stagsave;
      y = adjust(t);   // Anpassung an Ausgangsstagprec.
      y = y & einfachgenau;
  }

  return y;
}

l_interval sqrt(const l_interval &x, int n) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)       // NSqrt(n,x)
{
   int         stagsave = stagprec,
               stagshort, staglong, i,
               stagmax = 19,
               max = 2;
   l_interval  my, fy, corr,
               y;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = sqrt(dx,n);

   if (Inf(x) < 0.0) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval sqrt(const l_interval &x, int n)"));
   else if (stagprec == 1)
      y = pow(dx, _interval(1.0/n));
   else if (Inf(dx) == Sup(dx) && (Sup(dx) == CXSC_Zero || Sup(dx) == CXSC_One))
      y = x;
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
         
      while (max < stagprec) 
         max += max;  // quadratische Konvergenz kann
                      // nicht eingehalten werden,
      max += max;     // deshalb eine Schleife mehr

      y = pow(dx, _interval(1.0/n));

      stagprec = 1;

      // Intervall-Newton-Verfahren
      for (i = 2; i <= max; i += i) 
      {

         // Verdopplung der Genauigkeit
         stagshort = stagprec;
         stagprec += stagprec;
         if (stagprec > stagmax) 
            stagprec = stagmax;
         staglong  = stagprec;

         my = _l_interval(mid(y));
         fy = power(my, n)-x;

         // Fehlerauswertung
         stagprec  = stagshort;
         corr = fy/(_real(n)*power(y, n-1));
         stagprec  = staglong;

         // Fehlerkorrektur
         y = my-corr;
      }
      stagprec = stagsave;
      y = adjust(y);    // Anpassung an Ausgangsstagprec.
      y = y & einfachgenau;
   }

   return y;
}

l_interval sin(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)    // Sin(x)
{
   int         stagsave = stagprec,
               stagmax = 19;
   l_interval  pihalbe,
               y;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = sin(dx);

   if (stagprec == 1) 
      y = sin(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      
      pihalbe = 2.0*atan(_l_interval(1.0));
      y = x-pihalbe;
      
      try {
         y = cos(y);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval sin(const l_interval & x)")); 
      }
      
      
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval cos(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)   // Cos(x)
{
   long int    mm = 6;
   int         stagsave = stagprec,
               stagmax = 19,
               n = 0,
               digits = 53,
               sign = 0,
               degree, k,
               m2;
               
   bool        xinf = false,
               xsup = false,
               fertig = false;

   real        abst, m, eps,
//               bas, tn, t4,
               lneps, lnt,
               zk,
               zhn = 1.0,
               lnb = 0.69314718,
               fak = 720.0;    // 6!
   interval    dx = _interval(x),
               einfachgenau,
               dt, extr, error;
   l_interval  zwopi, t, t2, p, ph, test,
               y;

   einfachgenau = cos(dx);
   if (stagprec == 1) 
      y = cos(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      // stagprec++;
      zwopi = 8.0*atan(_l_interval(1.0));
      // stagprec--;

      // erste Argumentreduktion ( cos(x) = cos(x+2k*pi) )
      if (Sup(zwopi) < Sup(abs(x))) 
      {
         m = floor(_double(_real(Sup((x/zwopi)))));  // floor rundet zur nchsten
         t = x-m*zwopi;                              // ganzen Zahl kleiner x
         if (Sup(zwopi) < Sup(abs(t))) 
         { // das ganze nochmal, falls m bergelaufen ist!
                                      // ohne Sup wird Inclusion geprft!
            m = floor(_double(_real(Sup((t/zwopi)))));// rundet zur nchsten Zahl < x
            t = t-m*zwopi;
         }
      } else 
         t = x;

      // berprfen, ob Maximum oder Minimum im Inneren des Intervalls
      extr = _interval(2.0/zwopi*t);
      m2 = int(_double(floor(_double(Sup(extr)))));
      if (_interval(_real(m2)) <= extr) 
      {
         if (_interval(_real(m2-1),_real(m2)) <= extr) 
         {
            y = _l_interval(-1.0,1.0);
            fertig = true;
         } else 
            if (m2%2)  
               xinf = TRUE;
            else            
               xsup = TRUE;
      }

      if (!(fertig)) 
      {
         // zweite Argumentreduktion
         dt = _interval(t);
         // eps = 0.01/(stagprec*stagprec*stagprec);
         eps = 0.01;
         while (Sup(abs(dt))/zhn >= eps) 
         {
            n++;
            zhn += zhn;
         }
         t /= zhn;

         // Abschtzung der Genauigkeit

         t2 = t*t;
         abst = _real(Sup(abs(t)));
         if (abst < MinReal) 
            abst = MinReal; // nur zur Sicherheit
         lnt = ln(abst);
         lneps = (1.0-digits*stagprec)*lnb;
         while (lneps-(_real(mm)*lnt+ln(2.0/fak)) <= 0.0) 
         {
            mm  += 4;
            if (mm > 170) 
            {   
               // 170! = 7.26*10^306
               cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cos(const l_interval & x)"));
               mm = 170;
               break;
            }
            fak *= mm*(mm-1.0)*(mm-2.0)*(mm-3.0);
         }
         /*
            bas = _real(Inf(power(_l_interval(2.0),(1-digits*stagprec))));
            tn = abs(_real(Sup(power(t,6))));
            t4 = _real(mid(power(t,4)));
            while (bas-2.0*tn/fak <= 0.0) 
            {
               mm += 4;
               if (mm > 170) 
               {    // 170! = 7.26*10^306
                  errmon (ERROR_LINTERVAL(FAKOVERFLOW));
                  mm = 170;
                  break;
            }
            tn *= t4;
            fak *= mm*(mm-1.0)*(mm-2.0)*(mm-3.0);
         }
         */

         degree = mm-2;     // Achtung mm := 2n+2 !

         // Polynomauswertung

         sign = (degree/2)%2;
         zk  = _real(degree)*_real(degree-1);
         if (sign) 
            p = -t2/zk;
         else      
            p = t2/zk;
         for (k = degree-2; k >= 2; k -= 2) 
         {
            sign = 1-sign;
            if (sign)   
               p -= 1.0;
            else       
               p += 1.0;
            zk = _real(k)*_real(k-1);
            p *= t2/zk;
         }

         error = pow(_interval(2.0), _interval(1.0-digits*stagprec))
                 * _interval(-0.5,0.5);
         p = _l_interval(1.0)+error+p;

         // Rckgngigmachung der zweiten Argumentreduktion
         for (int i = 0; i < n; i++) 
            p = 2.0*p*p-1.0;

         stagprec = stagsave;
         y = adjust(p);
         if (Inf(y) < -1.0) 
            SetInf(y,-1.0);
         if (Sup(y) > 1.0) 
            SetSup(y,1.0);
         if (xinf) 
            SetInf(y,-1.0);
         if (xsup) 
            SetSup(y,1.0);
      }
      y = y & einfachgenau;
   }
   return y;
}

l_interval tan(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW,ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)   // Tan(x)
{
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  s, c, y;

   einfachgenau = tan(dx);

   if (stagprec == 1) 
      y = tan(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
         y = adjust(_l_interval(0.0));
   else 
   {
      try 
      {
         c = cos(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval tan(const l_interval &x)"));
      }

      if (_interval(0.0) <= c) 
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval tan(const l_interval &x)"));
      }
      try
      {
         s = sin(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval tan(const l_interval &x)"));
      }
      stagprec++;
      y = s/c;
      stagprec--;
      y = adjust(y);
      y = y & einfachgenau;
   }
   return y;
}

l_interval cot(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW,ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)   // Cot(x)
{
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  s, c, y;

   einfachgenau = cot(dx);

   if (stagprec == 1) 
      y = tan(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else 
   {
      try
      {
         s = sin(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cot(const l_interval &x)"));
      }

      if (_interval(0.0) <= s) 
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval cot(const l_interval &x)"));
      }

      try 
      {
         c = cos(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &) // Damit Funktionsname stimmt
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cot(const l_interval &x)"));
      }
   
      c = cos(x);
      stagprec++;
      y = c/s;
      stagprec--;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval asin(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)  // ASin(x)
{
   l_interval  t, ta, u, pihalbe,
               y;
   interval    dx = _interval(x),
               einfachgenau;
   real        supabsdx = Sup(abs(dx)),
               infdx = Inf(dx),
               supdx = Sup(dx);

   einfachgenau = asin(dx);

   stagprec++;
   pihalbe = 2.0*atan(_l_interval(1.0));
   stagprec--;
   
   if (Inf(x) < CXSC_MinusOne || Sup(x) > CXSC_One) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval asin(const l_interval & x)"));
   else if (stagprec == 1) 
      y = asin(dx);
   else if (infdx == supdx && supdx == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else if (infdx == supdx && supabsdx == CXSC_One) 
   {
      if (supdx == 1.0) 
         y = pihalbe;
      else              
         y = -pihalbe;
   } else 
   {
      stagprec++;
      try 
      {
         if (supabsdx <= 0.75) 
            u = x;
         else                  
            u = 2.0*x*sqrt((1.0-x)*(1.0+x));
         t = u/sqrt((1.0-u)*(1.0+u));
         stagprec--;
         ta = atan(t);
      }
      catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval asin(const l_interval & x)"));
      }
      stagprec++;
      if (supabsdx <= 0.75) 
         y = ta;
      else if (Sup(t) >= 0.0)  
         y = pihalbe-0.5*ta;
      else 
         y = -pihalbe-0.5*ta;
      stagprec--;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval acos(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)   // ACos(x)
{
   bool        neg=false;
   l_interval  pi, s, y;
   interval    dx = _interval(x),
               einfachgenau;
   real        supabsdx = Sup(abs(dx)),
               infdx = Inf(dx),
               supdx = Sup(dx);

   try {

      einfachgenau = acos(dx);

      stagprec++;
      pi = 4.0*atan(_l_interval(1.0));
      stagprec--;
      if (Inf(x) < CXSC_MinusOne || Sup(x) > CXSC_One) 
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF());
      else if (stagprec == 1) 
         y = acos(dx);
      else if (infdx == supdx && supabsdx == CXSC_One) 
      {
         if (supdx == 1.0) 
            y = adjust(_l_interval(0.0));
         else              
            y = adjust(pi);
      } else 
      {
         if (supdx < 0.0) 
         {
            y = -x;
            neg = true;
         } else 
         {
            y = x;
            neg = false;
         }
         stagprec++;
         if (supabsdx > 0.75) 
         {
            y = (1.0-(y))*(1.0+(y));
            //      stagprec--;
            y = asin(sqrt(y));
            //      stagprec++;
            if (neg)  
               y = pi-(y);
         } else 
         {
            //      stagprec--;
            s = asin(y);
            //      stagprec++;
            if (neg)  
               y = 0.5*pi+s;
            else      
               y = 0.5*pi-s;
         }

         // Fehler in der Systemumgebung, deshalb die Aufblhung des Intervalls
         real      err1, err2;
         interval  error;
         err1 = 5.0*_real(diam(y));
         err2 = 5.0*power(10.0,-16*(stagprec-1)-1);
         if (err1 > err2)  
            error = _interval(-err1, err1);
         else              
            error = _interval(-err2, err2);
         y += error;

         stagprec--;
         y = adjust(y);
         y = y & einfachgenau;
      }
   }
   catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
   {
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval acos(const l_interval & x)"));
   }
   return y;
}

static real    CXSC_pi4[21];
static bool    CXSC_pi4_initialized = false;


l_interval atan(const l_interval & x) throw()  // ATan(x)
{
   int         stagsave = stagprec,
               stagmax = 19,
               digits = 53,
               m = 1,
               stagsave2,
               degree, sign, zk;
   real        lnt2, lneps, ln3, ln3m, eps,
   //            bas, tn, t3,
               zhn = 1.0,
               lnb = 0.69314718;
   l_interval  t, t2, p,
               y;
   interval    dx = _interval(x),
               einfachgenau,
               err;

   einfachgenau = atan(dx);

   if (stagprec == 1) 
      y = atan(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(0.0));
   else 
   {
      if (dx == _interval(1.0)) 
      {
         stagmax = 20;
         if (!CXSC_pi4_initialized) 
         {
            std::string str;
            std::cout << SaveOpt;
            std::cout << Hex;
            str = "+1921FB54442D18e3FE";
            str >> CXSC_pi4[0];
            str = "+11A62633145C06e3C8";
            str >> CXSC_pi4[1];
            str = "+1C1CD129024E08e393";
            str >> CXSC_pi4[2];
            str = "+114CF98E804178e35E";
            str >> CXSC_pi4[3];
            str = "-159C4EC64DDAECe327";
            str >> CXSC_pi4[4];
            str = "+1410F31C6809BCe2F2";
            str >> CXSC_pi4[5];
            str = "-106AE64C32C5BCe2BB";
            str >> CXSC_pi4[6];
            str = "-1C99FA9EB241B4e286";
            str >> CXSC_pi4[7];
            str = "-1D791603D95252e24E";
            str >> CXSC_pi4[8];
            str = "-1571EDD0DBD254e218";
            str >> CXSC_pi4[9];
            str = "-133B4302721768e1E2";
            str >> CXSC_pi4[10];
            str = "+10BA698DFB5AC2e1AD";
            str >> CXSC_pi4[11];
            str = "+1FFAE5B7A035C0e178";
            str >> CXSC_pi4[12];
            str = "-1211C79404A576e143";
            str >> CXSC_pi4[13];
            str = "-1816945836FBA0e10D";
            str >> CXSC_pi4[14];
            str = "-1DA700CDB6BCCEe0D8";
            str >> CXSC_pi4[15];
            str = "+11ECE45B3DC200e0A3";
            str >> CXSC_pi4[16];
            str = "+1F2E2858EFC166e06D";
            str >> CXSC_pi4[17];
            str = "+1B4906C38ABA73e036";
            str >> CXSC_pi4[18];
            str = "+19A458FEA3F493e000";
            str >> CXSC_pi4[19];
            str = "+19A458FEA3F494e000";
            str >> CXSC_pi4[20];

            std::cout << RestoreOpt;

            CXSC_pi4_initialized = true;
         }
         stagprec = stagmax;
         y = adjust(_l_interval(0.0));        // y mit 20 real's anlegen
         for (int i = 0; i <= stagmax; i++)
            y.data[i] = CXSC_pi4[i];
         stagprec = stagsave;
         y = adjust(y);                // y mit stagprec real's
      } // Ende Spezialbehandlung
      else 
      {
         if (stagprec < stagmax) 
            stagprec++;
         else                    
            stagprec = stagmax;

         // Argumentreduktion
         t = x;
         eps = 0.01/stagprec;    // vorher 0.1, aber zu gro
         while (Sup(abs(_interval(t))) > eps) 
         {
            t = t/(1.0+sqrt(1.0+t*t));
            zhn += zhn;
         }
         t2 = t*t;

         // Bestimmung des Approximationsgrades

         lnt2 = Sup(ln(_interval(abs(t2))));
         ln3 = ln(3.0);
         ln3m = ln(3.0-_real(Sup(t2)));
         lneps = (1.0-digits*stagprec)*lnb;
         do {
            m += 3;
         } while (lneps-ln3-_real(m+1)*lnt2+ln(2.0*m+3.0)+ln3m <= 0.0);
         /*
            lnt = Sup(ln(_interval(abs(t))));     // bei Andreas u. Christian "t2"
            lneps = (1.0-digits*stagprec)*lnb;
            do {
               m += 3;
            } while (lneps+ln(2*m+3)-lnt*(2*m+3) <= 0.0);

            bas = _real(power(_l_real(2.0),1-digits*stagprec));
            tn = abs(_real(Sup(t)));
            t3 = abs(_real(Sup(power(t,3))));
            do {
               m += 3;
               tn *= t3;
               if (tn == 0.0) 
                  break; // bei Unterlauf erfolgt keine Fehlermeldung!
            } while (bas-tn/m <= 0.0);
         */
         degree = m;

         // Approximation
         sign = (degree%2) ? -1 : 1;
         zk = sign*(2*degree+1);
         p = _l_interval(1.0)/_real(zk);
         for (int k = degree-1; k >= 0; k--) 
         {
            sign = -sign;
            zk = sign*(2*k+1);
            p = p*t2+_l_interval(1.0)/_real(zk);
         } // letztes t wird beim Fehler mit einmultipliziert

         // Fehler bestimmen
         stagsave2 = stagprec;
         stagprec = 2;
         l_interval relerr;
         stagprec = stagsave2;
         err = pow(_interval(2.0), _interval(1.0-digits*stagprec))
                  * _interval(-1.0, 1.0);      // von 2^(2-d*s) gendert !
         relerr[1] = 1.0;
         relerr[2] = Inf(err);
         relerr[3] = Sup(err);

         // Argumentreduktion rckgngig machen, mit t mult. und Fehler einbauen
         y = zhn*t*p*relerr;

         stagprec = stagsave;
         y = adjust(y);
         y = y & einfachgenau;
      }
   }

   return y;
}

l_interval acot(const l_interval &x) throw()  // ACot(x)
{
   l_interval  pihalbe, y;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = acot(dx);
   stagprec++;
   pihalbe = 2.0*atan(_l_interval(1.0));
   stagprec--;

   if (stagprec == 1)
      y = acot(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(pihalbe);
   else 
   {
      y = pihalbe-atan(x);
      // y = atan(1.0/x);  andere Mglichkeit der Berechnung
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}


l_interval exp(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)   // Exp(x)
{
   long int    n = 2;
   int         stagsave = stagprec,
               stagmax = 19,
               degree,
               rednum = 0,
               digits = 53;
   real        fak = 2.0,
               zhn = 1.0,       // kann grer 32768 werden
               tmp, lny, lneps, eps,
               lnb = 0.69314718;
   l_interval  p, t,
               y;
   interval    dx = _interval(x),
               einfachgenau,
               dt, error;
   //            ilnb =_interval(0.6931471, 0.6931572),
   //            ilny, ilneps;

   einfachgenau = exp(dx);

   // gltigen Bereich prfen
   if (stagprec == 1) 
      y = exp(dx);
   else if (Inf(dx) == Sup(dx) && Inf(dx) == CXSC_Zero)
      y = adjust(_l_interval(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      if (Sup(dx) <= 0.0)  
         t = -x;  // Werte unter -1e308 nicht auswertbar,
      else                 
         t =  x;  // dafr andere Ergebnisse exakter
    
      dt = _interval(t);

      // Argumentreduktion
      //    eps = 0.000001*_real(Sup(t))/stagprec;
      //    eps = 1.0/pow10(stagprec/4);
      eps = 0.1;
      while(Sup(dt)/zhn > eps) 
      {
         rednum++;
         zhn += zhn;
      }
      t /= _l_interval(zhn);

      // Taylor Approximation
      dt  = _interval(t);

      // Anzahl der Durchlufe bei der Approximation
      tmp = Sup(abs(dt));
      if (MinReal > tmp) 
         tmp = MinReal;
      lny = ln(tmp);
      lneps = (1.0-digits*stagprec)*lnb;
      
      while(lneps-lnb+ln(fak)-_real(n)*lny <= 0.0) 
      {
         n += 3;
         if (n > 170) 
         {    // 170! = 7.26*10^306
            cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval exp(const l_interval & x)"));
            n = 170;
            break;
         }
         fak = fak*_real(n)*(n-1.0)*(n-2.0);
       }
      degree = n;

      p = t/_real(degree);
      int i;
      for (i=degree-1; i >= 1; i--)
         p = (p+1.0)*t/_real(i);

      // Fehlerabschtzung
      error = _interval(-2.0, 2.0)*pow(dt, _interval(_real(n)))/fak;

      // Fehler einbauen
      p += 1.0+_l_interval(error);

      // Argumentreduktion wird rckgngig gemacht
      for (i = 1; i <= rednum; i++)
         p *= p;

      if (Sup(dx) < 0.0)  
         p = 1.0/p;

      stagprec = stagsave;
      y = adjust(p);
      y = y & einfachgenau;
   }

   return y;
}

int cxsc_zweihoch(int) throw();

l_interval ln(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF)    // Ln(x)
{ 
   int         stagsave = stagprec,
               stagmax = 19,
               k,
               m = 0,
               n = 0,
               digits = 53,   // Anzahl der Mantissenstellen
               cmp,
               zhn1 = 1,
               zhn2 = 1;
   bool        zwei=false;
   real        mx, tmp, 
   //            zkp1,
               ln2 = 0.69314718,
               lnb = 0.69314718,
               lny, lneps;
   l_interval  y, t, t2, p;
   interval    dx = _interval(x),
               einfachgenau,
               dy, error;

   einfachgenau = ln(dx);

   if (Inf(x) <= 0.0) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval ln(const l_interval & x)"));
   else if (stagprec == 1) 
      y = ln(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_One)
      y = adjust(_l_interval(0.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      // erste Argumentreduktion
      /*
         if (Inf(dx) > 1.0) 
         {
            y = 1.0/x;
            dx = _interval(x);
            neg = true;
         } else 
            y = x;
      */
      y = x;

      // zweite Argumentreduktion
      mx = Sup(dx);
      tmp = 1.0/Sup(dx);
      if (tmp > mx) 
         mx=tmp;
      tmp = 1.0/Inf(dx);
      if (tmp > mx) 
         mx = tmp;
      if (mx > 1.1) 
      {
         cmp = int(_double(ln(ln(mx)/ln(1.1))/ln2));
         if (cmp > 0) 
         {
            if (cmp > 14) 
               n = 14;
            else 
               n = cmp;
            zhn1 = cxsc_zweihoch(n);
         }
      }
      y = sqrt(y, zhn1);
      mx = abs(_real(Sup(y)));
      if (mx > 1.1) 
      {
         cmp = int(_double(ln(ln(mx)/ln(1.1))/ln2));
         if (cmp > 0) 
         {
            if (cmp > 14) 
               n = 14;
            else 
               n = cmp;
            zhn2 = cxsc_zweihoch(n);
         }
         y = sqrt(y, zhn2);
         zwei = TRUE;
      }

      // dritte Argumentreduktion
      t = (y-1.0)/(y+1.0);

      // Abschtzung der Genauigkeit
      t2 = t*t;
      tmp = Sup(_interval(t2));
      if (tmp == 0.0) 
         tmp = MinReal;
      dy = tmp;
      lny = Sup(ln(dy));
      lneps = (1.0-digits*stagprec)*lnb;
      
      do {
         m += 2;
      } while (lneps-lnb+ln(2.0*m+3.0)-_real(m+1)*lny <= 0.0);

      // Polynomauswertung
      p = 0.0;
      for (k = 2*m+1; k >= 1; k -= 2)
         p = p*t2+_l_interval(2.0)/_real(k);
      p *= t;

      // Fehlereinschlieung
      error = _interval(-4.0, 4.0)*pow(_interval(t2), _interval(_real(m+1)))/(2.0*m+3.0);

      // Fehler einbauen
      p = p+error;

      // Argumentreduktion 2 wird rckgngig gemacht
      y = _real(zhn1)*p;
      if (zwei) 
         y *= _real(zhn2);

      // Argumentreduktion 1 wird rckgngig gemacht
      /*    if (neg) y = -(y);*/

      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;

      // Zustzliche Fehlerbehandlung ntig, da Underflow-Fehler nicht abgefangen werden kann
      /*    
         error = _interval(-4.0*pow10(-stagprec*16-1), 4.0*pow10(-stagprec*16-1));
         y += error;
      */

   }

   return y;
}

int cxsc_zweihoch(int n) throw()
{ 
   // liefert 2^n

   int res = 1;
   switch (n) 
   {
      case 0:
         return 1;
      case 1:
         return 2;
      case 2:
         return 4;
      default:
         {
            int y = 1,
            x = 2,
            zhi = 2;

            if (n%2)  
               res = 2;
            y = x*x;
            do {
               if ((n/zhi)%2)  
                  res *= y;
               zhi += zhi;
               if (zhi <= n)  
                  y *= y;
            } while (zhi <= n);
         }
   }
   return res;
}


l_interval sinh(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)   // Sinh(x)
{
   long int    n = 1;
   int         stagsave = stagprec,
               stagmax = 19,
               sign = 1,
               digits = 53,
               degree, stagsave2;
   real        tmp,
               lnt, lneps,
               fak = 1.0,
               lnb = 0.69314718;
   l_interval  y,
               t, t2, p, pot;
   interval    dy,
               dx = _interval(x),
               einfachgenau,
               ibase = _interval(2.0),
   //            ilnb = _interval(0.6931471,0.6931472),
   //            ilnx, ilneps, in,
               err;
   //            ifak = _interval(fak);

   einfachgenau = sinh(dx);

   if (stagprec == 1) 
      y = sinh(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = x;
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;

      // Ausnuztung der Punktsymmetrie
      if (Sup(dx) < 0.0) 
      {
         y = -x;
         sign = -1;
      } else  
         y =  x;
    
      dy = _interval(y);

      // Bei Werten ber 0.5 -- Auswertung ber e-Funktion
      if (Sup(dy) > 0.5) 
      {
         try {
            t = exp(y);
         }
         catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
         {
            cxscthrow( ERROR_LINTERVAL_FAK_OVERFLOW("l_interval sinh(const l_interval & x)")); 
         }
         y = sign*0.5*(t-1.0/t);
      }
      // Auswertung ber Potenzreihenentwicklung
      else 
      {
         t = y;
         t2 = t*t;

         // Abschtzung der Genauigkeit
         tmp = _real(Sup(abs(t)));       // bei Andreas und Christian "t2"
         if (tmp < MinReal)  
            tmp = MinReal;
         lnt = ln(tmp);
         lneps = (1.0-digits*stagprec)*lnb;
         do {
            n += 3;
            if (n > 170) 
            {    // 170! = 7.26*10^306
               cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval sinh(const l_interval & x)"));
               n = 170;
               fak *= n*(n-1.0);
               break;
            }
            fak *= n*(n-1.0)*(n-2.0);
         } while(lneps+ln(fak)-_real(n)*lnt-lnb <= 0.0);
         /*
            real bas, tn, t2d;
            bas = _double(_real(power(_l_real(2.0),1-digits*stagprec)));
            tn = abs(_double(_real(mid(t))));
            t2d = _double(_real(mid(t2)));
            do {
               n += 2.0;
               if (n > 170) 
               {    // 170! = 7.26*10^306
                  errmon(ERROR_LINTERVAL(FAKOVERFLOW));
                  n = 170;
                  fak *= n;
                  break;
               }
               tn *= t2d;
               fak *= n*(n-1.0);
            } while(bas-tn/fak <= 0.0);
         */
         degree = 2*(n/2+1);  // degree ist gerade
                              // Achtung degree = 2n+2!

         // Auswertung mit Horner-Schema
         p = 1.0;
         for (int i = degree; i >= 2; i -= 2) 
         {
            pot = _interval((i+1.0)*i);
            p = (p*(t2))/pot+1.0;
         }
         p *= t;

         // Negative Werte zurckwandeln
         if (sign == -1)  
            p = -p;

         // Fehlerauswertung
         stagsave2 = stagprec;
         stagprec = 2;
         l_interval  relerr;
         stagprec = stagsave2;
         err = pow(ibase , _interval(1.0-digits*stagprec)) * _interval(-1.0, 1.0);
         relerr[1] = 1.0;
         relerr[2] = Inf(err);
         relerr[3] = Sup(err);

         // Fehler einbauen
         y = p*relerr;
      }
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }
   return y;
}

l_interval cosh(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)   // Cosh(x)
{
   int         stagsave = stagprec,
               stagmax = 19;
   l_interval  y, t;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = cosh(dx);

   if (stagprec == 1) 
      y = cosh(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = adjust(_l_interval(1.0));
   else 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      if (Sup(dx) <= 0.0)  
         y = -x;
      else                 
         y =  x;
      
      try 
      {
         t = exp(y);
      } 
      catch(const  ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval cosh(const l_interval & x)"));
      }
      
      y = 0.5*(t+1.0/t);
      if (_interval(0.0) <= dx) 
         SetInf(y,1.0);
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval tanh(const l_interval & x) throw(ERROR_LINTERVAL_FAK_OVERFLOW)  // Tanh(x)
{
   interval    dx = _interval(x),
               einfachgenau;
   l_interval  e, s, c, y;

   einfachgenau = tanh(dx);
   if (stagprec == 1) 
      y = tanh(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = x;
   else 
   {
      try 
      {
         if (_interval(0.0) <= dx) 
         {
            c = cosh(x);
            s = sinh(x);
            stagprec++;
            y = s/c;
            stagprec--;
         } else 
         {
            stagprec++;
            e = exp(abs(x));
            if (Inf(x) > 0.0) 
               y = 1.0-2.0/(sqr(e)+1.0);
            else            
               y = -1.0+2.0/(sqr(e)+1.0);
            stagprec--;
         }
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval tanh(const l_interval & x)"));
      }

      if (Inf(y) < -1.0) 
         SetInf(y,-1.0);
      if (Sup(y) > 1.0)  
         SetSup(y,1.0);
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval coth(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW)  // Coth(x)           
{
   l_interval  s, c, y;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = coth(dx);

   if (stagprec == 1) 
      y = coth(dx);
   else if (_interval(0.0) <= dx) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval coth(const l_interval & x)"));
   else 
   {
      try 
      {
         c = cosh(x);
         s = sinh(x);
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval coth(const l_interval & x)"));
      }
      stagprec++;
      y = c/s;
      stagprec--;
      y = adjust(y);
      if (_interval(1.0) <= y) 
         SetInf(y,1.0);
      if (_interval(-1.0) <= y) 
         SetSup(y,-1.0);
      y = y & einfachgenau;
   }
   return y;
}
 
l_interval asinh(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW) // ASinh(x)
{
   int         stagsave = stagprec,
               stagmax = 19,
               stagsave2;
   l_interval  y, my;
   interval    dx = _interval(x),
               einfachgenau;
   //            error;

   try 
   {

      einfachgenau = asinh(dx);

      if (stagprec == 1) 
            y = asinh(dx);
      else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
            y = x;
      else if (abs(Sup(dx)) < 1e150) 
      {
         if (stagprec < stagmax) 
            stagprec++;
         else                    
            stagprec = stagmax;
         if (Sup(x) < 0.0) 
            y = -ln(-x+sqrt(x*x+1.0)); // keine Auslschung !
         else 
            y = ln(x+sqrt(x*x+1.0));                // Ausnutzung der Symmetrie
         stagprec = stagsave;
         y = adjust(y);

         // Zustzliche Fehlerbehandlung ntig, da Underflow-Fehler nicht abgefangen werden kann
         /*    
            error = _interval(-4.0*pow10(-stagprec*16-1), 4.0*pow10(-stagprec*16-1));
            y += error;
         */
      } else 
      {
         if (stagprec < stagmax) 
            stagsave2 = stagprec+1;
         else                    
            stagsave2 = stagmax;
         y = asinh(dx);
         stagprec = 1;
         do {
            stagprec += stagprec;
            if (stagprec > stagsave2) 
               stagprec = stagsave2;
            my = mid(y);
            y = my-(sinh(my)-x)/cosh(y);
         } while(stagprec < stagsave2);
         stagprec = stagsave;
         y = adjust(y);
         y = y & einfachgenau;
      }
   } 
   catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
   {
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval asinh(const l_interval & x)"));
   }
   catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
   {
      cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval asinh(const l_interval & x)"));
   }
   return y;
}

l_interval acosh(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW)  // ACosh(x)
{
   int         stagsave = stagprec,
               stagmax = 19,
               stagsave2;
   interval    dx = _interval(x),
               einfachgenau;
   //            error;
   l_interval  y, my;

   try
   {

      einfachgenau = acosh(dx);
      
      if (Inf(x) < 1.0) 
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF());
      else if (stagprec == 1)
         y = acosh(dx);
      else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_One)
         y = adjust(_l_interval(0.0));
      else if (Sup(dx) < 1e150) 
      {
         if (stagprec < stagmax) 
            stagprec++;
         else                    
            stagprec = stagmax;
         if (Sup(dx) < 0.0)  
            y = ln(x-sqrt(x*x-1.0));
         else                
            y = ln(x+sqrt(x*x-1.0));
         stagprec = stagsave;
         y = adjust(y);

         // Zustzliche Fehlerbehandlung ntig, da Underflow-Fehler nicht abgefangen werden kann
         /*    
            error = _interval(-4.0*pow10(-stagprec*16-1), 4.0*pow10(-stagprec*16-1));
            y += error;
         */
      } else 
      {
         if (stagprec < stagmax) 
            stagsave2 = stagprec+1;
         else                    
            stagsave2 = stagmax;
         y = acosh(dx);
         stagprec = 1;
         do {
            stagprec += stagprec;
            if (stagprec > stagsave2) 
               stagprec = stagsave2;
            my = mid(y);
            y = my-(cosh(my)-x)/sinh(y);
         } while(stagprec < stagsave2);
         stagprec = stagsave;
         y = adjust(y);
         if (Inf(x) == 1.0) 
            SetInf(y,0.0);
         y = y & einfachgenau;
      }

   }
   catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
   {
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval acosh(const l_interval & x)"));
   }
   catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
   {
      cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval acosh(const l_interval & x)"));
   }
   
   return y;
}


l_interval atanh(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW) // ATanh(x)
{
   int         stagsave = stagprec,
               stagmax = 19,
               stagsave2;
   l_interval  y, my;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = atanh(dx);

   if (Inf(x) <= CXSC_MinusOne || Sup(x) >= CXSC_One) 
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval atanh(const l_interval & x)"));
   else if (stagprec == 1)
      y = atanh(dx);
   else if (Inf(dx) == Sup(dx) && Sup(dx) == CXSC_Zero)
      y = x;
   else if (abs(Sup(dx)) < 1e150) 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      
      y = 0.5*ln((1.0+x)/(1.0-x));
      stagprec = stagsave;
      y = adjust(y);
   } else 
   {
      if (stagprec < stagmax) 
         stagsave2 = stagprec+1;
      else                    
         stagsave2 = stagmax;
      y = atanh(dx);
      stagprec = 1;
      try {
         do {
            stagprec += stagprec;
            if (stagprec > stagsave2) 
               stagprec = stagsave2;
            my = mid(y);
            y = my-(tanh(my)-x)*sqr(cosh(y));
         } while(stagprec < stagsave2);
      }
      catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval atanh(const l_interval & x)"));
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval atanh(const l_interval & x)"));
      }
      
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

l_interval acoth(const l_interval & x) throw(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF,ERROR_LINTERVAL_FAK_OVERFLOW) // ACoth(x)
{
   int         stagsave = stagprec,
               stagmax = 19,
               stagsave2;
   l_interval  y, my;
   interval    dx = _interval(x),
               einfachgenau;

   einfachgenau = acoth(dx);

   if ((_l_interval(Inf(x))) < _l_interval(CXSC_MinusOne,CXSC_One)
    || (_l_interval(Sup(x))) < _l_interval(CXSC_MinusOne,CXSC_One))
      cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval acoth(const l_interval & x)"));
   else if (stagprec == 1)
      y = acoth(dx);
   else if (abs(Sup(dx)) < 1e150) 
   {
      if (stagprec < stagmax) 
         stagprec++;
      else                    
         stagprec = stagmax;
      
      y = 0.5*ln((x+1.0)/(x-1.0));
      stagprec = stagsave;
      y = adjust(y);
   } else 
   {
      if (stagprec < stagmax) 
         stagsave2 = stagprec+1;
      else                    
         stagsave2 = stagmax;
     
      y = acoth(dx);
      stagprec = 1;
      try {
         do {
            stagprec += stagprec;
            if (stagprec > stagsave2) 
               stagprec = stagsave2;
            my = mid(y);
            y = my+(cosh(my)-x)*sqr(sinh(y));
         } while(stagprec < stagsave2);
      }
      catch(const ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF &)
      {
         cxscthrow(ERROR_LINTERVAL_STD_FKT_OUT_OF_DEF("l_interval acoth(const l_interval & x)"));
      }
      catch(const ERROR_LINTERVAL_FAK_OVERFLOW &)
      {
         cxscthrow(ERROR_LINTERVAL_FAK_OVERFLOW("l_interval acoth(const l_interval & x)"));
      }
         
      stagprec = stagsave;
      y = adjust(y);
      y = y & einfachgenau;
   }

   return y;
}

} // namespace cxsc
