/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: l_real.cpp,v 1.3 2001/08/09 11:37:57 cxsc Exp $ */

#include <memory.h>

#include "l_real.hpp"
#include "dot.hpp"
#include "idot.hpp"
#include "interval.hpp"
#include "l_interv.hpp"

namespace cxsc {

int stagprec=2;

interval::interval(const l_real & a) throw()
{
   dotakku[0]=a;
   rnd(dotakku[0],inf,sup);
}

interval & interval::operator =(const l_real & a) throw()
{
   dotakku[0]=a;
   rnd(dotakku[0],inf,sup);
   return *this;
}

idotprecision::idotprecision(const l_real & a) throw() : inf(0),
                                                         sup(0)
{ 
   *this+=a;
}

idotprecision::idotprecision(const l_real & a,const l_real & b) : 
                                                         inf(0),
                                                         sup(0)
{ 
   inf+=a;
   sup+=b;
}



void l_real::_clear(int p) throw()
{  
   // fuellt l_real ab Stelle p bis zum Ende mit Null.
   // funktioniert immer, da for-Schleife abweisend!
   for (int i=p; i<=prec; i++)
      this->elem(i)=0.0;
}


void l_real::_akku_out() throw()
{   
    // ein im dotakku[0] liegendes Zwischenergebnis wird in der entsprechenden
    // Praezision in die aufrufende l_real Zahl gerundet.
    _clear(1);
    this->elem(1) = rnd(dotakku[0]);
    if (prec>1) 
    {
       int i=2, weiter;
       do {
           dotakku[0] -= this->elem(i-1);
           this->elem(i) = rnd(dotakku[0]);
           i++;
           weiter = abs(this->elem(i-1)) >= MinReal;
       } while (weiter && (i <= prec));

       // Gendert am 23.09.92 von F. Toussaint
       // Abbruch war bei berprfung auf 0!
       if (abs(this->elem(i-1)) < MinReal) 
          this->elem(i-1) = 0.0;
    }
}

void l_real::_akku_add(dotprecision& d) const throw()
{ 
   // addiert aufrufenden l_real auf akku d.
   for (int i=1; i<=prec; i++) 
   {
      if (abs(this->elem(i)) >= MinReal) // Gendert am 23.09.92 von F. Toussaint
         d += this->elem(i);             // bisher wurde nur 0 abgefragt
                                         // im Unterlauf fehlerhafte Berechnung
   }
}

void l_real::_akku_sub(dotprecision& d) const throw()
{ 
   // subtrahiert aufrufenden l_real von akku d.
   for (int i=1; i<=prec; i++) 
   {
      if (abs(this->elem(i)) >= MinReal) // Gendert am 23.09.92 von F. Toussaint
         d -= this->elem(i);
   }
}

//---------------------------------------------------------------------------
// Konstruktoren und Destruktoren
//
l_real::l_real() throw()
{
   data=new real[stagprec];      
   if(data)
      prec=stagprec;
   else
      prec=0;
}

l_real::~l_real() throw()
{
   delete [] data;
}

l_real::l_real(const l_real& lr) throw() 
                                 : data(new real[lr.prec])
{
   if(data)
   {
      prec=lr.prec;
      memcpy(data,lr.data,sizeof(real)*prec);
   } else
      prec=0;
}

l_real::l_real(const real & r) throw()
                             : prec(1), 
                               data(new real[1])
{
   data[0]=r;
}

l_real::l_real(int i) throw()
                    : prec(1),
                      data(new real[1])
{
   data[0]=i;
}

//---------------------------------------------------------------------------
// Typtransfer

real::real(const l_real& lr) throw()
{
   dotakku[0]=0.0;
   lr._akku_add(dotakku[0]);
   w=rnd(dotakku[0]).w;
}

dotprecision::dotprecision(const l_real& lr) throw() : akku(new a_btyp[A_LENGTH])
{
   memset(akku,0,BUFFERSIZE);
   lr._akku_add(*this);
}

dotprecision & dotprecision::operator =(const l_real & lr) throw()
{
   memset(akku,0,BUFFERSIZE);
   lr._akku_add(*this);
   return *this;
}

l_real::l_real(const dotprecision & d) throw() : prec(stagprec),
                                                 data(new real[prec])
{
   dotakku[0]=d;
   _akku_out();
}

//---------------------------------------------------------------------------
// Zuweisungen
//
l_real& l_real::operator=(const real& r) throw()
{ 
   // Siehe _l_real(real..)

   if(prec!=1)
   {
      delete [] data;
      data=new real[1];
      if(data)
         prec=1;
      else
         prec=0;
   }
   elem(1)=r;
   return *this;
}

l_real& l_real::operator=(const l_real& lr) throw()
{ 
   // Siehe c++-FAQ assignment-operators.html#[12.1]

   // This code gracefully (albeit implicitly) handles self assignment
   real* tmp = new real[lr.prec];   // It would be OK if an exception got thrown here

   if((prec=lr.prec)>0)
      memcpy(tmp,lr.data,sizeof(real)*prec);

   delete [] data;
   data = tmp;
   return *this;
}

l_real& l_real::operator=(const dotprecision & d) throw()
{ 
   // Siehe _l_real(real..)

   if(prec!=stagprec)
   {
      delete [] data;
      data=new real[prec=stagprec];
   }
   dotakku[0]=d;
   _akku_out();   
   return *this;
}


real& l_real::operator[](int i) const throw() 
{
        if (i<1 || i>prec) 
        {
                // throw!
                i=1;
        }
        return data[i-1];
}

int StagPrec(const l_real& lr) throw()
{
        return lr.prec;
}

std::istream& operator>>(std::istream& s, l_real& lr) throw()
{
        s >> dotakku[0];
        lr._akku_out();
        return s;
}

std::ostream& operator<<(std::ostream& s, const l_real& lr) throw()
{
        dotakku[0]=0.0;
        lr._akku_add(dotakku[0]);
        s << dotakku[0];
        return s;
}
std::string& operator>>(std::string& s, l_real& lr) throw()
{
        s >> dotakku[0];
        lr._akku_out();
        return s;
}

std::string& operator<<(std::string& s, const l_real& lr) throw()
{
        dotakku[0]=0.0;
        lr._akku_add(dotakku[0]);
        s << dotakku[0];
        return s;
}
void operator>>(const std::string& s, l_real& lr) throw()
{
   std::string r(s);
   r>>lr;
}
void operator>>(const char *s, l_real& lr) throw()
{
   std::string r(s);
   r>>lr;
}


void accumulate(dotprecision& d, const real& r, const l_real& lr) throw()
{
   accumulate(d, l_real(r), lr);
}

void accumulate(dotprecision& d, const l_real& lr, const real& r) throw()
{
   accumulate(d, lr, l_real(r));
}

void accumulate(dotprecision& d, const l_real& lr1, const l_real&lr2) throw()
{
   // nderung am 24.9.92 von F. Toussaint wegen Underflow-Fehler
   int i, j;

   for (i=1; i<=lr1.prec; i++)
      for (j=1; j<=lr2.prec; j++)
         if (abs(lr1.elem(i) * lr2.elem(j)) > MinReal) // <----------- ?????
            accumulate(d, lr1.elem(i), lr2.elem(j));
}

void accumulate(idotprecision & a, const real & b, const l_real & c) throw() { accumulate(a,l_interval(b),l_interval(c)); }
void accumulate(idotprecision & a, const l_real & b, const real & c) throw() { accumulate(a,l_interval(b),l_interval(c)); }
void accumulate(idotprecision & a, const l_real & b, const l_real & c) throw() { accumulate(a,l_interval(b),l_interval(c)); }



l_real  operator-(const l_real& lr, const real& r) throw() { return lr-_l_real(r); }
l_real  operator-(const real& r, const l_real& lr) throw() { return _l_real(r)-lr; }
l_real  operator+(const l_real& lr, const real& r) throw() { return lr+_l_real(r); }
l_real  operator+(const real& r, const l_real& lr) throw() { return _l_real(r)+lr; }
l_real  operator*(const l_real& lr, const real& r) throw() { return lr*_l_real(r); }
l_real  operator*(const real& r, const l_real& lr) throw() { return _l_real(r)*lr; }
l_real  operator/(const l_real& lr, const real& r) throw() { return lr/_l_real(r); }
l_real  operator/(const real& r, const l_real& lr) throw() { return _l_real(r)/lr; }

dotprecision operator-(const l_real& lr, const dotprecision& r) throw() 
{ 
   return _dotprecision(lr)-r; 
}
dotprecision operator-(const dotprecision& r, const l_real& lr) throw() 
{ 
   return r-_dotprecision(lr); 
}
dotprecision operator+(const l_real& lr, const dotprecision& r) throw() 
{ 
   return _dotprecision(lr)+r; 
}
dotprecision operator+(const dotprecision& r, const l_real& lr) throw() 
{ 
   return r+_dotprecision(lr); 
}


l_real& operator-=(l_real& lr, const real& r) throw()
{ lr = lr-_l_real(r); return lr; }
    
l_real& operator+=(l_real& lr, const real& r) throw()
{ lr = lr+_l_real(r); return lr; }

l_real& operator*=(l_real& lr, const real& r) throw()
{ lr = lr*_l_real(r); return lr; }

l_real& operator/=(l_real& lr, const real& r) throw()
{ lr = lr/_l_real(r); return lr; }

real& operator-=(real& r, const l_real& lr) throw() { r = r-_real(lr); return r; }
real& operator+=(real& r, const l_real& lr) throw() { r = r+_real(lr); return r; }
real& operator*=(real& r, const l_real& lr) throw() { r = r*_real(lr); return r; }
real& operator/=(real& r, const l_real& lr) throw() { r = r/_real(lr); return r; }

bool operator==(const l_real& lr, const real& r) throw() { return lr==_l_real(r); }
bool operator==(const real& r, const l_real& lr) throw() { return _l_real(r)==lr; }
bool operator!=(const l_real& lr, const real& r) throw() { return lr!=_l_real(r); }
bool operator!=(const real& r, const l_real& lr) throw() { return _l_real(r)!=lr; }

bool operator<=(const l_real& lr, const real& r) throw() { return lr<=_l_real(r); }
bool operator<=(const real& r, const l_real& lr) throw() { return _l_real(r)<=lr; }
bool operator>=(const l_real& lr, const real& r) throw() { return lr>=_l_real(r); }
bool operator>=(const real& r, const l_real& lr) throw() { return _l_real(r)>=lr; }

bool operator<(const l_real& lr, const real& r) throw() { return lr<_l_real(r); }
bool operator<(const real& r, const l_real& lr) throw() { return _l_real(r)<lr; }
bool operator>(const l_real& lr, const real& r) throw() { return lr>_l_real(r); }
bool operator>(const real& r, const l_real& lr) throw() { return _l_real(r)>lr; }

bool operator==(const l_real& lr, const dotprecision& d) throw() { return _dotprecision(lr)==d; }
bool operator==(const dotprecision& d, const l_real& lr) throw() { return d==_dotprecision(lr); }
bool operator!=(const l_real& lr, const dotprecision& d) throw() { return _dotprecision(lr)!=d; }
bool operator!=(const dotprecision& d, const l_real& lr) throw() { return d!=_dotprecision(lr); }

bool operator<=(const l_real& lr, const dotprecision& d) throw() { return _dotprecision(lr)<=d; }
bool operator<=(const dotprecision& d, const l_real& lr) throw() { return d<=_dotprecision(lr); }
bool operator>=(const l_real& lr, const dotprecision& d) throw() { return _dotprecision(lr)>=d; }
bool operator>=(const dotprecision& d, const l_real& lr) throw() { return d>=_dotprecision(lr); }

bool operator<(const l_real& lr, const dotprecision& d) throw() { return _dotprecision(lr)<d; }
bool operator<(const dotprecision& d, const l_real& lr) throw() { return d<_dotprecision(lr); }
bool operator>(const l_real& lr, const dotprecision& d) throw() { return _dotprecision(lr)>d; }
bool operator>(const dotprecision& d, const l_real& lr) throw() { return d>_dotprecision(lr); }

l_real operator-(const l_real& lr1) throw()
{
   l_real lr2(lr1);
   for (int i=1; i<=lr1.prec; i++)
      lr2.elem(i) =  -(lr1.elem(i));
   return lr2;
}
l_real operator+(const l_real& lr1) throw()
{
   return lr1;
}

l_real operator-(const l_real& lr1, const l_real& lr2) throw()
{
   l_real lr3;
   dotakku[0]=0.0;
   lr1._akku_add(dotakku[0]);
   lr2._akku_sub(dotakku[0]);
   lr3._akku_out();
   return lr3;
}

l_real operator+(const l_real& lr1, const l_real& lr2) throw()
{
   l_real lr3;
   dotakku[0]=0.0;
   lr1._akku_add(dotakku[0]);
   lr2._akku_add(dotakku[0]);
   lr3._akku_out();

   return lr3;
}

l_real operator*(const l_real& lr1, const l_real& lr2) throw()
{
   l_real lr3;
   dotakku[0]=0.0;
   accumulate(dotakku[0], lr1, lr2);
   lr3._akku_out();
   return lr3;
}

l_real operator/(const l_real& lr1, const l_real& lr2) throw()
{
   real a, b;
   l_real lr3;
   lr3._clear(1);
   dotakku[0]=0.0;
   dotakku[1]=0.0;
   lr1._akku_add(dotakku[0]);
   lr2._akku_add(dotakku[1]);
   a = rnd(dotakku[0], RND_DOWN);
   b = rnd(dotakku[1], RND_UP);
   if (!b) 
   {
      // errmon (ERR_LREAL(DIVZERO));
      // throw!
   } else 
   {
      lr3.elem(1) = a/b;
      for (int i=2; i<=stagprec; i++) 
      {
         if (!a) 
            break;
         for (int j=1; j<=lr2.prec; j++)
            if (!!lr3.elem(i-1) && !!lr2.elem(j) )
               accumulate(dotakku[0], lr3.elem(i-1), -lr2.elem(j));
         a = rnd(dotakku[0], RND_DOWN);
         lr3.elem(i) = a/b;
      }
   } // if keine Division durch Null
   return lr3;
}


l_real & operator-=(l_real& lr1, const l_real& lr2) throw() { return lr1 = lr1-lr2; }
l_real & operator+=(l_real& lr1, const l_real& lr2) throw() { return lr1 = lr1+lr2; }
l_real & operator*=(l_real& lr1, const l_real& lr2) throw() { return lr1 = lr1*lr2; }
l_real & operator/=(l_real& lr1, const l_real& lr2) throw() { return lr1 = lr1/lr2; }

bool operator==(const l_real& lr1, const l_real& lr2) throw()
{
   dotakku[0]=0.0;
   dotakku[1]=0.0;
   lr1._akku_add(dotakku[0]);
   lr2._akku_add(dotakku[1]);
   return (dotakku[0]==dotakku[1]);
}

bool operator<=(const l_real& lr1, const l_real& lr2) throw()
{
   dotakku[0]=0.0;
   dotakku[1]=0.0;
   lr1._akku_add(dotakku[0]);
   lr2._akku_add(dotakku[1]);
   return (dotakku[0]<=dotakku[1]);
}

bool operator!(const l_real& lr) throw()
{
   dotakku[0]=0.0;
   lr._akku_add(dotakku[0]);
   return (!dotakku[0]);
}

bool operator!=(const l_real& lr1, const l_real& lr2) throw() { return !(lr1==lr2); }
bool operator< (const l_real& lr1, const l_real& lr2) throw() { return !(lr2<=lr1); }
bool operator> (const l_real& lr1, const l_real& lr2) throw() { return !(lr1<=lr2); }
bool operator>=(const l_real& lr1, const l_real& lr2) throw() { return (lr2<=lr1);  }

// ID-LR

 idotprecision operator +(const idotprecision &a,const l_real &b) throw() 
{ 
   return idotprecision(a.inf+b,a.sup+b); 
}

 idotprecision operator +(const l_real &b,const idotprecision &a) throw() 
{ 
   return idotprecision(a.inf+b,a.sup+b); 
} 

 idotprecision operator -(const idotprecision &a,const l_real &b) throw() 
{ 
   return idotprecision(a.inf-b,a.sup-b); 
} 

 idotprecision operator -(const l_real &a,const idotprecision &b) throw() 
{ 
   return idotprecision(a-b.sup,a-b.inf); 
}

idotprecision operator |(const l_real &a,const idotprecision &b) throw() 
{
   return idotprecision((a<b.inf)?_dotprecision(a):b.inf,(a>b.sup)?_dotprecision(a):b.sup);
}

 idotprecision operator |(const idotprecision &a,const l_real &b) throw() 
{
   return idotprecision((a.inf<b)?a.inf:_dotprecision(b),(a.sup>b)?a.sup:_dotprecision(b));
}

 idotprecision operator &(const l_real &a,const idotprecision &b) throw(ERROR_IDOTPRECISION_EMPTY_INTERVAL) 
{
   return idotprecision((a>b.inf)?_dotprecision(a):b.inf,(a<b.sup)?_dotprecision(a):b.sup);
}

 idotprecision operator &(const idotprecision &a,const l_real &b) throw(ERROR_IDOTPRECISION_EMPTY_INTERVAL) 
{
   return idotprecision((a.inf>b)?a.inf:_dotprecision(b),(a.sup<b)?a.sup:_dotprecision(b));
}

// LR-ID
idotprecision & operator +=(idotprecision &a,const l_real &b) throw() { return a+=dotprecision(b); }      
idotprecision & operator -=(idotprecision &a,const l_real &b) throw() { return a-=dotprecision(b); }
idotprecision & operator |=(idotprecision &a,const l_real &b) throw() { return a|=dotprecision(b); }
idotprecision & operator &=(idotprecision &a,const l_real &b) throw(ERROR_IDOTPRECISION_EMPTY_INTERVAL) { return a&=dotprecision(b); }


bool operator ==(const interval &i,const l_real &r) throw() { return Inf(i)==r && Sup(i)==r; }
bool operator !=(const interval &i,const l_real &r) throw() { return Inf(i)!=r || Sup(i)!=r; }
bool operator  <(const interval &i,const l_real &r) throw() { return false; }
bool operator  >(const interval &i,const l_real &r) throw() { return Inf(i)<r && Sup(i)>r; }
bool operator <=(const interval &i,const l_real &r) throw() { return Inf(i)==r && Sup(i)==r; }
bool operator >=(const interval &i,const l_real &r) throw() { return Inf(i)<=r && Sup(i)>=r; }

bool operator ==(const l_real &r,const interval &i) throw() { return Inf(i)==r && Sup(i)==r; }
bool operator !=(const l_real &r,const interval &i) throw()   { return Inf(i)!=r || Sup(i)!=r; }
bool operator  <(const l_real &r,const interval &i) throw()   { return Inf(i)<r && Sup(i)>r; }
bool operator  >(const l_real &r,const interval &i) throw()   { return false; }
bool operator <=(const l_real &r,const interval &i) throw()   { return Inf(i)<=r && Sup(i)>=r; }
bool operator >=(const l_real &r,const interval &i) throw()   { return Inf(i)==r && Sup(i)==r; }

bool operator ==(const idotprecision &i,const l_real &r) throw() { return Inf(i)==r && Sup(i)==r; }
bool operator !=(const idotprecision &i,const l_real &r) throw() { return Inf(i)!=r || Sup(i)!=r; }
bool operator  <(const idotprecision &i,const l_real &r) throw() { return false; }
bool operator  >(const idotprecision &i,const l_real &r) throw() { return Inf(i)<r && Sup(i)>r; }
bool operator <=(const idotprecision &i,const l_real &r) throw() { return Inf(i)==r && Sup(i)==r; }
bool operator >=(const idotprecision &i,const l_real &r) throw() { return Inf(i)<=r && Sup(i)>=r; }

bool operator ==(const l_real &r,const idotprecision &i) throw() { return Inf(i)==r && Sup(i)==r; }
bool operator !=(const l_real &r,const idotprecision &i) throw()   { return Inf(i)!=r || Sup(i)!=r; }
bool operator  <(const l_real &r,const idotprecision &i) throw()   { return Inf(i)<r && Sup(i)>r; }
bool operator  >(const l_real &r,const idotprecision &i) throw()   { return false; }
bool operator <=(const l_real &r,const idotprecision &i) throw()   { return Inf(i)<=r && Sup(i)>=r; }
bool operator >=(const l_real &r,const idotprecision &i) throw()   { return Inf(i)==r && Sup(i)==r; }



//---------------------------------------------------------------------------
// Sonstige Funktionen
//
l_real  abs(const l_real& lr1) throw()
{
   l_real lr2;
   dotakku[0]=0.0;
   lr1._akku_add(dotakku[0]);

   if (dotakku[0] < 0.0) 
      dotakku[0] = -dotakku[0];

   lr2._akku_out();

   return lr2;
}

int sign(const l_real& lr) throw()
{
   dotakku[0]=0.0;
   lr._akku_add(dotakku[0]);
   return sign(dotakku[0]);
}

l_real adjust(const l_real & x) throw()
{
   l_real  y;
  
   if (x.prec == stagprec) 
      y = x;
   else if (x.prec > stagprec) 
   {
      dotakku[0]=0.0;
      x._akku_add(dotakku[0]);
      y._akku_out();
   } else 
   {
      int i;
      for (i = 0; i <= stagprec-x.prec-1; i++)
         y.data[i] = 0;
      for (i = stagprec-x.prec; i <= stagprec-1; i++)
         y.data[i] = x.data[i-(stagprec-x.prec)];
   }
        
   return y;
}

} // namespace cxsc
